%
% This experiment considers an application of the achromatic morphological gradient to detect the boundaries
% 	of an object in a natural color image (caps).
%
% This experimet uses some codes provided by Peter Kovesi
% 	Centre for Exploration Targeting
% 	The University of Western Australia
% 	Available at: http://www.peterkovesi.com/matlabfns/
%
% See the paper: 
%	M.E. Valle and R.A. Valente Neto. 
% 	Mathematical Morphology on the Spherical CIELab Quantale for Colour Image Processing, 
% 	Submitted for publication.
%

clear;

% Read the color image "caps".
ImA = imread('caps.png');
% Convert the image to double.
ImA = im2double(ImA);

% Read the ground truth image;
GT=~im2bw(imread('yellow_cap.jpg'));

% Show the image and the ground truth image;
figure(1); imshow(ImA);
title('Natural color image');

figure(2); imshow(1-GT);
title('Ground truth Image');

% Define the reference color as yellow;
ref=[1,1,0];

% Construct the non-planar structuring element.
s1=1.2; 
s(:,:,1)=[s1 s1 s1; s1 1 s1; s1 s1 s1];
s(:,:,2)=(pi/2)*ones(3,3); 
s(:,:,3)=pi*ones(3,3);

% Construc the planar structuring element.
p(:,:,1)=ones(3,3);
p(:,:,2)=(pi/2)*ones(3,3);
p(:,:,3)=pi*ones(3,3);

% Detine the hystersis threshold;
T1=0.01;
T2=0.2;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Spherical CIELab approach based on a flat SE;

disp('Spherical CIELab approach using a flat');

% Compute the flat morphological gradient.
G=SphCIELabMM(ImA,ref,p,'grad');

% Re-scale the gradient image;
G=G(:,:,1)/max(max(G(:,:,1)));

% Show the re-scaled gradient image;
figure(3); imshow(1-G);
title('Spherical CIELab gradient image based on a flat SE');

% Non-maxima suppression;
orientim = featureorient(G,0,1);
GM=nonmaxsup(G,orientim,1.5);

% Hystheresis threshold
B=hysthresh(GM,T1,T2);

% Show the binary image;
figure(4); imshow(1-B);
title('Boundary image obtained using a flat SE');

% Compute Pratt's Figure of Merit;
FOM=PrattFOM(B,GT)

disp('');
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Spherical CIELab approach based on a non-flat SE;

disp('Spherical CIELab approach using a non-flat');

% Compute the flat morphological gradient.
G=SphCIELabMM(ImA,ref,s,'grad');

% Re-scale the gradient image;
G=G(:,:,1)/max(max(G(:,:,1)));

% Show the re-scaled gradient image;
figure(5); imshow(1-G);
title('Spherical CIELab gradient image based on a non-flat SE');

% Non-maxima suppression;
orientim = featureorient(G,0,1);
GM=nonmaxsup(G,orientim,1.5);

% Hystheresis threshold
B=hysthresh(GM,T1,T2);

% Show the binary image;
figure(6); imshow(1-B);
title('Boundary image obtained using a non-flat SE');

% Compute Pratt's Figure of Merit;
FOM=PrattFOM(B,GT)