%
% This experiment illustrate the morphological operations on the spherical CIELab quantale using a synthetic image.
%
% See the paper: 
%	M.E. Valle and R.A. Valente Neto. 
% 	Mathematical Morphology on the Spherical CIELab Quantale for Colour Image Processing, 
% 	Submitted for publication.
%

clear;

% The following constructs the synthetic image;
% Define a color map;
map=[0.5 0.5 0.5; 0 0 0; 0 0 1; 1 0 1; 1 0 0; 0 1 0; 0 1 1; 1 1 1; 1 1 0];
	
% Construct a simple color image;
indmap=0;
for i=1:3
	for j=1:3
		indmap=indmap+1;
		indi=3*(i-1)+1;
		indj=3*(j-1)+1;
		for xi=1:3
			ImA(indi:indi+2,indj:indj+2,xi)=map(indmap,xi)*ones(3,3);
		end
	end
end		

% Show the synthetic color image;
figure(1); clf;
imagesc(ImA); grid;
title('Original Color Image');

% Define the color reference as a pure red;
ref=[1,0,0];

% Construct the non-planar structuring element.
s1=2;
s(:,:,1)=[s1 s1 s1; s1 1 s1; s1 s1 s1];
s(:,:,2)=(pi/2)*ones(3,3); 
s(:,:,3)=pi*ones(3,3);

% Construc the planar structuring element.
p(:,:,1)=ones(3,3);
p(:,:,2)=(pi/2)*ones(3,3);
p(:,:,3)=pi*ones(3,3);

% Compute the flat dilation, erosion, and achromatic morphological gradient of the color image ImA. 
[ImRGB_Ero,ImSph_Ero]=SphCIELabMM(ImA,ref,p,'ero'); figure(2); clf; imagesc(ImRGB_Ero); title('Flat Erosion');
[ImRGB_Dil,ImSph_Dil]=SphCIELabMM(ImA,ref,p,'dil'); figure(3); clf; imagesc(ImRGB_Dil); title('Flat Dilation');
[ImRGB_Grad,ImSph_Grad]=SphCIELabMM(ImA,ref,p,'grad'); figure(4); clf; imagesc(1-ImRGB_Grad/max(max(ImRGB_Grad(:,:,1)))); title('Flat Gradient');

% Compute the non-flat dilation, erosion, and achromatic morphological gradient of the color image ImA. 
[ImRGB_Ero,ImSph_Ero]=SphCIELabMM(ImA,ref,s,'ero'); figure(5); clf; imagesc(ImRGB_Ero); title('Non-Flat Erosion');
[ImRGB_Dil,ImSph_Dil]=SphCIELabMM(ImA,ref,s,'dil'); figure(6); clf; imagesc(ImRGB_Dil); title('Non-Flat Dilation');
[ImRGB_Grad,ImSph_Grad]=SphCIELabMM(ImA,ref,s,'grad'); figure(7); clf; imagesc(1-ImRGB_Grad/max(max(ImRGB_Grad(:,:,1)))); title('Non-Flat Gradient');