#include "mex.h"
#include <stdio.h>

void mexFunction(int nlhs, mxArray *plhs[], 
    int nrhs, const mxArray *prhs[])

{
 double aux;
 int M, N;
 if (nrhs != 3)
 	mexErrMsgTxt("The number of input arguments must be three.");

 /* Find the dimensions of the input matrices; */
 M = mxGetM(prhs[0]);
 N = mxGetM(prhs[1]);
 
 if (mxGetN(prhs[0]) != 2)
	mexErrMsgTxt("Error: The first input must contain the row and column indices of the actual image.");
  
 if (mxGetN(prhs[1]) != 2)
	mexErrMsgTxt("Error: The second input must contain the row and column indices of the ground truth image.");
 
 /* Create an mxScalar for the output data */
 plhs[0] = mxCreateDoubleMatrix(1, 1, mxREAL);

 int i,j;
 double *A, *B, *output, *alpha;
 double dist, min_dist;
 
 /* Retrieve the input data */
 A = mxGetPr(prhs[0]);
 B = mxGetPr(prhs[1]);
 alpha = mxGetPr(prhs[2]);

 /* Create a pointer to the output data */
 output = mxGetPr(plhs[0]);

 /* Compute Pratt's Figure of Merit */
 output[0]=0;
 for (i=0; i<M; i++)
 {
   min_dist=(A[i]-B[0])*((A[i]-B[0])) + (A[i+M]-B[N])*(A[i+M]-B[N]);
   j=1;
   while ((j<N)&&(min_dist>0))
   {
     dist=(A[i]-B[j])*((A[i]-B[j])) + (A[i+M]-B[j+N])*(A[i+M]-B[j+N]);
     if (dist<min_dist)
     {
       min_dist = dist;
     }
     j++;
   }
   output[0]=output[0]+1/(1+alpha[0]*min_dist);
   /*    printf("alpha=%2.2f. min_dist=%2.2f. output=%2.2f \n",alpha[0],min_dist,output[0]); */
 }
 if (M<N)
 {
 /*  printf("output=%2.4f/%d \n",output[0],N); */
   output[0]=output[0]/N;
 }
 else
 {
 /*    printf("output=%2.4f/%d \n",output[0],M); */
   output[0]=output[0]/M;
 }
}
     
