function [ImRGB, ImSph] = SphCIELabMM(A, color, SE, operator)
%
% Morphological operations on the spherical CIELab quantale.
%
% Structure: [ImRGB, ImSph] = SphCIELabMM(ImA, reference, SE, operator)
%
% Input:  	ImA = color image in the RGB color system.
%		reference = reference color vector in the RGB color system.
%		SE = structuring element in the RGB color system.
%		operator = desired morphological operator:
%	  	'dil'		Dilation
%  		'ero'		Erosion
%  		'grad'		Gradient
%  		'open'		Opening
%  		'clos'		Closing
%
% Output: 	ImRGB = image in the RGB color system.
%		ImSph = image in the spherical CIELab space.
%
% Remark: 
%  	ImRGB, ImSph and A have the same size.
%  	
%  	If no SE is given, SphCIELabMM returns the image ImA in RGB (ImRGB) and spherical CIELab systems (ImSph).
%  
% Reference paper: 
%	M.E. Valle and R.A. Valente Neto. 
% 	Mathematical Morphology on the Spherical CIELab Quantale for Colour Image Processing, 
% 	Submitted for publication.
%
%  Copyright (c) 2016 Marcos Eduardo Valle and Raul Ambrozio Valente
%  valle@ime.unicamp.br
%  
%  Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
%  and associated documentation files (the "Software"), to deal in the Software without restriction, 
%  including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
%  and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
%  subject to the following conditions:
%  
%  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
%  
%  The software is provided "as is", without warranty of any kind, express or implied, 
%  including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. 
%  In no event shall the authors or copyright holders be liable for any claim, damages or other liability, 
%  whether in an action of contract, tort or otherwise, arising from, out of or in connection with 
%  the software or the use or other dealings in the software. 
%
%

if exist('SphCIELabMMmex')~=3
  mex SphCIELabMMmex.c;
end

if nargin<3
  [ImRGB, ImSph] = SphCIELabMMmex(A, color);
  disp('No SE was given. The outputs are the image in RGB and spherical CIELab system, respectivelly.');
else
  if nargin<4
    disp('Please, inform the operation: dil (dilation), ero (erosion), grad (gradient), open (opening), and clos (closing).');
  else
    [ImRGB, ImSph] = SphCIELabMMmex(A, color, SE, operator);
  end
end


  