/*
Copyright (c) 2016 Marcos Eduardo Valle and Raul Ambrozio Valente
valle@ime.unicamp.br

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

The software is provided "as is", without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the software. 
*/

#include "mex.h"
#include "string.h"
#include "math.h"
#include "stdio.h"

#define wx 0.950456
#define wy 1
#define wz 1.088754

#define BOTX 100000
#define BOTY 0
#define BOTZ 0

#define PI 3.141592653589793

#define EX 0
#define EY (PI/2)
#define EZ PI

int less(double x, double y, int n)
{
  if (n == 0)
  {
    if ((fabs(x)<fabs(y))||((fabs(x)==fabs(y))&&(x<y)))
      return 1;
    else
      return 0;
  }
  else
  {
    if (x==y)
      return 1;
    else
      if ((fabs(x)<fabs(y))||((fabs(x)==fabs(y))&&(x<y)))
        return 1;
      else
        return 0;
  }
}


double divextended(double y, double x)
{
  if (x >= BOTX)
    return 0;
  else if ((x == 0)&&(y == 0))
    return 0;
  else if ((x == 0)&&(y > 0))
    return BOTX;
  else
    return (y/x);
}

double divtheta(double y, double x)
{
  if (less(x,y,1))
    return PI;
  else
    return y;
}


double divphi(double y, double x)
{
  if (less(x,y,1))
    return (PI/2);
  else
    return y;
}

double inf(double x, double y)
{
    if (less(x,y,1))
      return x;
    else
      return y;
}


double fnl(double t)
{
    if (t > 0.008856)
      return (pow(t,1.0/3.0));
    else
      return (7.787*t+((double) 16.0/116.0));
}

double fnlinv(double t)
{
    if (t > 0.206893034422964)
      return (t*t*t);
    else
      return ((t-((double) 16.00/116.00))/7.787);
}


float min(float a, float b)
{
  if (a<b)
    return a;
  else
    return b;
}


float max(float a, float b)
{
  if (a>b)
    return a;
  else
    return b;
}

void mexFunction(int nlhs, mxArray *plhs[], 
    int nrhs, const mxArray *prhs[])
{
  double *vecR, *Ref, *A, *RGBOUT, *SESphCIELab, *SphCIELabOUT;
  const mwSize *size, *sizeSE;
  char *input;
  int buflen, status;
  
  buflen = mxGetN(prhs[nrhs-1])*sizeof(mxChar)+1;
  input = mxMalloc(buflen);
  status = mxGetString(prhs[nrhs-1], input, buflen);

  /* ################################################################# */
  /* Test Input Parameters */
  /* ----------------------------------------------------------------- */
  
  if (nrhs < 2)
        mexErrMsgTxt("Number of input arguments must be at least 2.");

  /* ################################################################# */
  /* Common Operations */
  /* ----------------------------------------------------------------- */
  
  /* Get Size of Input Image (Matrix)*/
  size = mxGetDimensions(prhs[0]);
  
  /* Get the matrix of input Image A in RGB */
  A = mxGetPr(prhs[0]);
  
  /* Get the vector vecR of Color reference*/
  vecR = mxGetPr(prhs[1]);
  Ref = mxGetPr(mxCreateDoubleMatrix(1, 3, mxREAL));
  
  /* Create an mxArray for the output image in RGB */
  plhs[0] = mxCreateNumericArray(size[2], size, mxDOUBLE_CLASS, mxREAL);
  RGBOUT = mxGetPr(plhs[0]);
  
  /* Create an mxArray for the output image in SphCIELab */
  plhs[1] = mxCreateNumericArray(size[2], size, mxDOUBLE_CLASS, mxREAL);
  SphCIELabOUT = mxGetPr(plhs[1]);
  
  int i, j, x, y, z;
  double fnlx, fnly, fnlz, Deltax, Deltay, Deltaz;
  
  /* Convert vector in RGB to vector in SphCIELab */  
  fnlx = fnl((0.412453*vecR[0]+0.357580*vecR[1]+0.180423*vecR[2])/wx);
  fnly = fnl((0.212671*vecR[0]+0.715160*vecR[1]+0.072169*vecR[2])/wy);
  fnlz = fnl((0.019334*vecR[0]+0.119193*vecR[1]+0.950227*vecR[2])/wz);
  Ref[0] = (116.00*fnly-16.00);
  Ref[1] = (500.00*(fnlx-fnly));
  Ref[2] = (200.00*(fnly-fnlz));
  
  /* ################################################################# */
  /* No Struct Element */
  /* ----------------------------------------------------------------- */
  
  if (nrhs == 2)
  {
    /* Convert image A in RGB to SphCIELab */
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	fnlx = fnl((0.412453*A[x]+0.357580*A[y]+0.180423*A[z])/wx);
	fnly = fnl((0.212671*A[x]+0.715160*A[y]+0.072169*A[z])/wy);
	fnlz = fnl((0.019334*A[x]+0.119193*A[y]+0.950227*A[z])/wz);
	Deltax = (116.00*fnly-16.00-Ref[0]);
	Deltay = (500.00*(fnlx-fnly)-Ref[1]);
	Deltaz = (200.00*(fnly-fnlz)-Ref[2]);
	SphCIELabOUT[x] = sqrt(Deltax*Deltax+Deltay*Deltay+Deltaz*Deltaz);
	SphCIELabOUT[y] = atan2(Deltax,sqrt(Deltay*Deltay+Deltaz*Deltaz));
	SphCIELabOUT[z] = atan2(Deltaz,Deltay);
	RGBOUT[x] = A[x];
	RGBOUT[y] = A[y];
	RGBOUT[z] = A[z];
      }
    }    
    return;
  }
  
  /* ################################################################# */
  /* Dilation */
  
  if (strcmp(input,"dil") == 0)
  {
    printf("Spherical CIELab Quantale-Based Dilation.\n");
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	fnlx = fnl((0.412453*A[x]+0.357580*A[y]+0.180423*A[z])/wx);
	fnly = fnl((0.212671*A[x]+0.715160*A[y]+0.072169*A[z])/wy);
	fnlz = fnl((0.019334*A[x]+0.119193*A[y]+0.950227*A[z])/wz);
	Deltax = (116.00*fnly-16.00-Ref[0]);
	Deltay = (500.00*(fnlx-fnly)-Ref[1]);
	Deltaz = (200.00*(fnly-fnlz)-Ref[2]);
	/* SphCIELabOUT will be the least element */
	SphCIELabOUT[x] = BOTX;
	SphCIELabOUT[y] = BOTY;
	SphCIELabOUT[z] = BOTZ;
	/* RGBOUT will be the image A in Spherical Coordenates */
	RGBOUT[x] = sqrt(Deltax*Deltax+Deltay*Deltay+Deltaz*Deltaz); /* r */
	RGBOUT[y] = atan2(Deltax,sqrt(Deltay*Deltay+Deltaz*Deltaz)); /* phi */
	RGBOUT[z] = atan2(Deltaz,Deltay); /* theta */
      }
    }
    /* Get Size of Input Matrix */
    sizeSE = mxGetDimensions(prhs[2]);

    /* Get the matrix of input SE */
    SESphCIELab = mxGetPr(prhs[2]);

    int k, l, xSE, ySE, zSE, limX, limY, xORIG, yORIG, zORIG;
    double mult, infy, infz;

    /* Limits of SE */
    limX = floor(sizeSE[0]/2);
    limY = floor(sizeSE[1]/2);

    for (k = 0; k < size[0]; k++)
    {
      for (l = 0; l < size[1]; l++)
      {
	for (i = 0; i < sizeSE[0]; i++)
	{
	  for (j = 0; j < sizeSE[1]; j++)
	  {
	    /* Test if SE fits in image A */
	    if (((k+i-limX)>=0)&&((k+i-limX)<size[0])&&((l+j-limY)>=0)&&((l+j-limY)<size[1]))
	    {
	      xSE = i+j*sizeSE[0];
	      ySE = i+j*sizeSE[0]+sizeSE[0]*sizeSE[1];
	      zSE = i+j*sizeSE[0]+2*sizeSE[0]*sizeSE[1];
	      Deltax = k + (i-limX);
	      Deltay = l + (j-limY);
	      xORIG = l*size[0]+k;
	      yORIG = l*size[0]+k+size[0]*size[1];
	      zORIG = l*size[0]+k+2*size[0]*size[1];
	      x = Deltay*size[0]+Deltax;
	      y = Deltay*size[0]+Deltax+size[0]*size[1];
	      z = Deltay*size[0]+Deltax+2*size[0]*size[1];
	      
	      /* Note that if RGBOUT[x] or SESphCIELab[xSE] are
	       * greater or equal to BOTX, then SphCIELabOUT[xORIG]
	       * cannot be greater than it. */
	      if ((RGBOUT[x]>=BOTX)||(SESphCIELab[xSE]>=BOTX))
		mult = BOTX;
	      else
	      {
		mult = RGBOUT[x]*SESphCIELab[xSE];
		infy = inf(RGBOUT[y],SESphCIELab[ySE]);
		infz = inf(RGBOUT[z],SESphCIELab[zSE]);

		/* Test the greast value between SphCIELabOUT and RGBOUT */
		if ((mult<SphCIELabOUT[xORIG])||((mult==SphCIELabOUT[xORIG])&&(less(SphCIELabOUT[yORIG],infy, 0)))||((mult==SphCIELabOUT[xORIG])&&(SphCIELabOUT[yORIG]==infy)&&(less(SphCIELabOUT[zORIG],infz, 1))))
		{
		    SphCIELabOUT[xORIG] = mult;
		    SphCIELabOUT[yORIG] = infy;
		    SphCIELabOUT[zORIG] = infz;
		}
	      }
	    }
	  }
	}
      }
    }
  
    /* Convert image in SphCIELab to RGB*/
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	
	fnlx = SphCIELabOUT[x]*sin(SphCIELabOUT[y])+Ref[0];
	fnly = SphCIELabOUT[x]*cos(SphCIELabOUT[y])*cos(SphCIELabOUT[z])+Ref[1];
	fnlz = SphCIELabOUT[x]*cos(SphCIELabOUT[y])*sin(SphCIELabOUT[z])+Ref[2];
	Deltax = wx*(fnlinv((fnlx+16.00)/116.00+fnly/500.00));
	Deltay = wy*fnlinv((fnlx+16.00)/116.00);
	Deltaz = wz*(fnlinv((fnlx+16.00)/116.00-fnlz/200.00));
	RGBOUT[x] = min(1,max(0,(+3.240479*Deltax-1.537150*Deltay-0.498535*Deltaz)));
	RGBOUT[y] = min(1,max(0,(-0.969256*Deltax+1.875992*Deltay+0.041556*Deltaz)));
	RGBOUT[z] = min(1,max(0,(+0.055648*Deltax-0.204043*Deltay+1.057311*Deltaz)));
      }
    }
  }
  /* ################################################################# */
  /* Erosion */
  /* ----------------------------------------------------------------- */

  else if(strcmp(input,"ero") == 0)
  {
    printf("Spherical CIELab Quantale-Based Erosion.\n");
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	fnlx = fnl((0.412453*A[x]+0.357580*A[y]+0.180423*A[z])/wx);
	fnly = fnl((0.212671*A[x]+0.715160*A[y]+0.072169*A[z])/wy);
	fnlz = fnl((0.019334*A[x]+0.119193*A[y]+0.950227*A[z])/wz);
	Deltax = (116.00*fnly-16.00-Ref[0]);
	Deltay = (500.00*(fnlx-fnly)-Ref[1]);
	Deltaz = (200.00*(fnly-fnlz)-Ref[2]);
	/* SphCIELabOUT will be the greast element */
	SphCIELabOUT[x] = EX;
	SphCIELabOUT[y] = EY;
	SphCIELabOUT[z] = EZ;
	/* RGBOUT will be the image A in Spherical Coordenates */
	RGBOUT[x] = sqrt(Deltax*Deltax+Deltay*Deltay+Deltaz*Deltaz); /* r */
	RGBOUT[y] = atan2(Deltax,sqrt(Deltay*Deltay+Deltaz*Deltaz)); /* phi */
	RGBOUT[z] = atan2(Deltaz,Deltay); /* theta */
      }
    }
    
    /* Struct Element (SE) */
	
    /* Get Size of Input Matrix */
    sizeSE = mxGetDimensions(prhs[2]);

    /* Get the matrix of input SE */
    SESphCIELab = mxGetPr(prhs[2]);

    int k, l, xSE, ySE, zSE, limX, limY, xORIG, yORIG, zORIG;
    double divx, divy, divz;

    /* Limits of SE */
    limX = floor(sizeSE[0]/2);
    limY = floor(sizeSE[1]/2);

    for (k = 0; k < size[0]; k++)
    {
      for (l = 0; l < size[1]; l++)
      {
	for (i = 0; i < sizeSE[0]; i++)
	{
	  for (j = 0; j < sizeSE[1]; j++)
	  {
	    /* Test if SE fits in image A */
	    if (((k+i-limX)>=0)&&((k+i-limX)<size[0])&&((l+j-limY)>=0)&&((l+j-limY)<size[1]))
	    {
	      xSE = i+j*sizeSE[0];
	      ySE = i+j*sizeSE[0]+sizeSE[0]*sizeSE[1];
	      zSE = i+j*sizeSE[0]+2*sizeSE[0]*sizeSE[1];
	      Deltax = k + (i-limX);
	      Deltay = l + (j-limY);
	      xORIG = l*size[0]+k;
	      yORIG = l*size[0]+k+size[0]*size[1];
	      zORIG = l*size[0]+k+2*size[0]*size[1];
	      x = Deltay*size[0]+Deltax;
	      y = Deltay*size[0]+Deltax+size[0]*size[1];
	      z = Deltay*size[0]+Deltax+2*size[0]*size[1];
	      
	      divx = divextended(RGBOUT[x],SESphCIELab[xSE]);
	      divy = divphi(RGBOUT[y],SESphCIELab[ySE]);
	      divz = divtheta(RGBOUT[z],SESphCIELab[zSE]);
	     
	      if ((divx>SphCIELabOUT[xORIG])||((divx==SphCIELabOUT[xORIG])&&(less(divy, SphCIELabOUT[yORIG], 0)))||((divx==SphCIELabOUT[xORIG])&&(SphCIELabOUT[yORIG]==divy)&&(less(divz, SphCIELabOUT[zORIG], 1))))
	      {
		SphCIELabOUT[xORIG] = divx;
		SphCIELabOUT[yORIG] = divy;
		SphCIELabOUT[zORIG] = divz;
	      }
	    }
	  }
	}
      }
    }
    
    /* Convert image in SphCIELab to RGB*/
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	
	fnlx = SphCIELabOUT[x]*sin(SphCIELabOUT[y])+Ref[0];
	fnly = SphCIELabOUT[x]*cos(SphCIELabOUT[y])*cos(SphCIELabOUT[z])+Ref[1];
	fnlz = SphCIELabOUT[x]*cos(SphCIELabOUT[y])*sin(SphCIELabOUT[z])+Ref[2];
	Deltax = wx*(fnlinv((fnlx+16.00)/116.00+fnly/500.00));
	Deltay = wy*fnlinv((fnlx+16.00)/116.00);
	Deltaz = wz*(fnlinv((fnlx+16.00)/116.00-fnlz/200.00));
	RGBOUT[x] = min(1,max(0,(+3.240479*Deltax-1.537150*Deltay-0.498535*Deltaz)));
	RGBOUT[y] = min(1,max(0,(-0.969256*Deltax+1.875992*Deltay+0.041556*Deltaz)));
	RGBOUT[z] = min(1,max(0,(+0.055648*Deltax-0.204043*Deltay+1.057311*Deltaz)));
      }
    }
  }
  
  /* ################################################################# */ 
  /* Gradient */
  /* ----------------------------------------------------------------- */
  
  else if(strcmp(input,"grad") == 0)
  {
    printf("Spherical CIELab Quantale-Based Achromatic Morphological Gradient.\n");
    double *B, *C;
    B = mxGetPr(mxCreateNumericArray(size[2], size, mxDOUBLE_CLASS, mxREAL));
    C = mxGetPr(mxCreateNumericArray(size[2], size, mxDOUBLE_CLASS, mxREAL));
    
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	fnlx = fnl((0.412453*A[x]+0.357580*A[y]+0.180423*A[z])/wx);
	fnly = fnl((0.212671*A[x]+0.715160*A[y]+0.072169*A[z])/wy);
	fnlz = fnl((0.019334*A[x]+0.119193*A[y]+0.950227*A[z])/wz);
	Deltax = (116.00*fnly-16.00-Ref[0]);
	Deltay = (500.00*(fnlx-fnly)-Ref[1]);
	Deltaz = (200.00*(fnly-fnlz)-Ref[2]);
	/* C will be the least element for Dilation */
	C[x] = BOTX;
	C[y] = BOTY;
	C[z] = BOTZ;
	/* B will be the identity element for Erosion */
	B[x] = EX;
	B[y] = EY;
	B[z] = EZ;
	/* RGBOUT will be the image A in Spherical Coordenates */
	RGBOUT[x] = sqrt(Deltax*Deltax+Deltay*Deltay+Deltaz*Deltaz); /* r */
	RGBOUT[y] = atan2(Deltax,sqrt(Deltay*Deltay+Deltaz*Deltaz)); /* phi */
	RGBOUT[z] = atan2(Deltaz,Deltay); /* theta */
      }
    }
    /* Get Size of Input Matrix */
    sizeSE = mxGetDimensions(prhs[2]);

    /* Get the matrix of input SE */
    SESphCIELab = mxGetPr(prhs[2]);

    int k, l, xSE, ySE, zSE, limX, limY, xORIG, yORIG, zORIG;
    double mult, infy, infz, divx, divy, divz;

    /* Limits of SE */
    limX = floor(sizeSE[0]/2);
    limY = floor(sizeSE[1]/2);

    /* Dilation and Erosion */
    for (k = 0; k < size[0]; k++)
    {
      for (l = 0; l < size[1]; l++)
      {
	for (i = 0; i < sizeSE[0]; i++)
	{
	  for (j = 0; j < sizeSE[1]; j++)
	  {
	    /* Test if SE fits in image A */
	    if (((k+i-limX)>=0)&&((k+i-limX)<size[0])&&((l+j-limY)>=0)&&((l+j-limY)<size[1]))
	    {
	      xSE = i+j*sizeSE[0];
	      ySE = i+j*sizeSE[0]+sizeSE[0]*sizeSE[1];
	      zSE = i+j*sizeSE[0]+2*sizeSE[0]*sizeSE[1];
	      Deltax = k + (i-limX);
	      Deltay = l + (j-limY);
	      xORIG = l*size[0]+k;
	      yORIG = l*size[0]+k+size[0]*size[1];
	      zORIG = l*size[0]+k+2*size[0]*size[1];
	      x = Deltay*size[0]+Deltax;
	      y = Deltay*size[0]+Deltax+size[0]*size[1];
	      z = Deltay*size[0]+Deltax+2*size[0]*size[1];
	      
	      /* Dilation */
	      /* Note that if RGBOUT[x] or SESphCIELab[xSE] are
	       * greater or equal to BOTX, then SphCIELabOUT[xORIG]
	       * cannot be greater than it. */
	      if ((RGBOUT[x]>=BOTX)||(SESphCIELab[xSE]>=BOTX))
		mult = BOTX;
	      else
	      {
		mult = RGBOUT[x]*SESphCIELab[xSE];
		infy = inf(RGBOUT[y],SESphCIELab[ySE]);
		infz = inf(RGBOUT[z],SESphCIELab[zSE]);

		/* Test the greast value between SphCIELabOUT and RGBOUT */
		if ((mult<C[xORIG])||((mult==C[xORIG])&&(less(C[yORIG],infy, 0)))||((mult==C[xORIG])&&(C[yORIG]==infy)&&(less(C[zORIG],infz, 1))))
		{
		    C[xORIG] = mult;
		    C[yORIG] = infy;
		    C[zORIG] = infz;
		}
	      }
	      
	      /* Erosion */
	      divx = divextended(RGBOUT[x],SESphCIELab[xSE]);
	      divy = divphi(RGBOUT[y],SESphCIELab[ySE]);
	      divz = divtheta(RGBOUT[z],SESphCIELab[zSE]);
	     
	      if ((divx>B[xORIG])||((divx==B[xORIG])&&(less(divy, B[yORIG], 0)))||((divx==B[xORIG])&&(B[yORIG]==divy)&&(less(divz, B[zORIG], 1))))
	      {
		B[xORIG] = divx;
		B[yORIG] = divy;
		B[zORIG] = divz;
	      }
	    }
	  }
	}
      }
    }
    
    /* Achromatic Gradient */
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	
	/* Compute the pixel Euclidean distance between the dilation and the erosion in the CIELab space */
	fnlx = (C[x]*sin(C[y])+Ref[0]) - (B[x]*sin(B[y])+Ref[0]);
	fnly = (C[x]*cos(C[y])*cos(C[z])+Ref[1]) - (B[x]*cos(B[y])*cos(B[z])+Ref[1]);
	fnlz = (C[x]*cos(C[y])*sin(C[z])+Ref[2]) - (B[x]*cos(B[y])*sin(B[z])+Ref[2]);
	Deltax = sqrt(fnlx*fnlx + fnly*fnly + fnlz*fnlz);
	RGBOUT[x] = Deltax;
	RGBOUT[y] = Deltax;
	RGBOUT[z] = Deltax;
      }
    }
  }
   
  /* ################################################################# */
  /* Closing */
  /* ----------------------------------------------------------------- */
  
  else if(strcmp(input,"clos") == 0)
  {
    printf("Spherical CIELab Quantale-Based Closing.\n");
    double *B;
    /* Get the matrix of input Image B in RGB */
    B = mxGetPr(mxCreateNumericArray(size[2], size, mxDOUBLE_CLASS, mxREAL));
    
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	fnlx = fnl((0.412453*A[x]+0.357580*A[y]+0.180423*A[z])/wx);
	fnly = fnl((0.212671*A[x]+0.715160*A[y]+0.072169*A[z])/wy);
	fnlz = fnl((0.019334*A[x]+0.119193*A[y]+0.950227*A[z])/wz);
	Deltax = (116.00*fnly-16.00-Ref[0]);
	Deltay = (500.00*(fnlx-fnly)-Ref[1]);
	Deltaz = (200.00*(fnly-fnlz)-Ref[2]);
	/* B will be the least element for Dilation */
	B[x] = BOTX;
	B[y] = BOTY;
	B[z] = BOTZ;
	/* SphCIELabOUT will be the identity element for Erosion */
	SphCIELabOUT[x] = EX;
	SphCIELabOUT[y] = EY;
	SphCIELabOUT[z] = EZ;
	/* RGBOUT will be the image A in Spherical Coordenates */
	RGBOUT[x] = sqrt(Deltax*Deltax+Deltay*Deltay+Deltaz*Deltaz); /* r */
	RGBOUT[y] = atan2(Deltax,sqrt(Deltay*Deltay+Deltaz*Deltaz)); /* phi */
	RGBOUT[z] = atan2(Deltaz,Deltay); /* theta */
      }
    }
    /* Get Size of Input Matrix */
    sizeSE = mxGetDimensions(prhs[2]);

    /* Get the matrix of input SE */
    SESphCIELab = mxGetPr(prhs[2]);

    int k, l, xSE, ySE, zSE, limX, limY, xORIG, yORIG, zORIG;
    double mult, infy, infz, divx, divy, divz;

    /* Limits of SE */
    limX = floor(sizeSE[0]/2);
    limY = floor(sizeSE[1]/2);

    /* Dilation */
    for (k = 0; k < size[0]; k++)
    {
      for (l = 0; l < size[1]; l++)
      {
	for (i = 0; i < sizeSE[0]; i++)
	{
	  for (j = 0; j < sizeSE[1]; j++)
	  {
	    /* Test if SE fits in image A */
	    if (((k+i-limX)>=0)&&((k+i-limX)<size[0])&&((l+j-limY)>=0)&&((l+j-limY)<size[1]))
	    {
	      xSE = i+j*sizeSE[0];
	      ySE = i+j*sizeSE[0]+sizeSE[0]*sizeSE[1];
	      zSE = i+j*sizeSE[0]+2*sizeSE[0]*sizeSE[1];
	      Deltax = k + (i-limX);
	      Deltay = l + (j-limY);
	      xORIG = l*size[0]+k;
	      yORIG = l*size[0]+k+size[0]*size[1];
	      zORIG = l*size[0]+k+2*size[0]*size[1];
	      x = Deltay*size[0]+Deltax;
	      y = Deltay*size[0]+Deltax+size[0]*size[1];
	      z = Deltay*size[0]+Deltax+2*size[0]*size[1];
	      
	      /* Dilation */
	      /* Note that if RGBOUT[x] or SESphCIELab[xSE] are
	       * greater or equal to BOTX, then SphCIELabOUT[xORIG]
	       * cannot be greater than it. */
	      if ((RGBOUT[x]>=BOTX)||(SESphCIELab[xSE]>=BOTX))
		mult = BOTX;
	      else
	      {
		mult = RGBOUT[x]*SESphCIELab[xSE];
		infy = inf(RGBOUT[y],SESphCIELab[ySE]);
		infz = inf(RGBOUT[z],SESphCIELab[zSE]);

		/* Test the greast value between SphCIELabOUT and RGBOUT */
		if ((mult<B[xORIG])||((mult==B[xORIG])&&(less(B[yORIG],infy, 0)))||((mult==B[xORIG])&&(B[yORIG]==infy)&&(less(B[zORIG],infz, 1))))
		{
		    B[xORIG] = mult;
		    B[yORIG] = infy;
		    B[zORIG] = infz;
		}
	      }
	    }
	  }
	}
      }
    }
    
    /* Erosion */
    for (k = 0; k < size[0]; k++)
    {
      for (l = 0; l < size[1]; l++)
      {
	for (i = 0; i < sizeSE[0]; i++)
	{
	  for (j = 0; j < sizeSE[1]; j++)
	  {
	    /* Test if SE fits in image A */
	    if (((k+i-limX)>=0)&&((k+i-limX)<size[0])&&((l+j-limY)>=0)&&((l+j-limY)<size[1]))
	    {
	      xSE = i+j*sizeSE[0];
	      ySE = i+j*sizeSE[0]+sizeSE[0]*sizeSE[1];
	      zSE = i+j*sizeSE[0]+2*sizeSE[0]*sizeSE[1];
	      Deltax = k + (i-limX);
	      Deltay = l + (j-limY);
	      xORIG = l*size[0]+k;
	      yORIG = l*size[0]+k+size[0]*size[1];
	      zORIG = l*size[0]+k+2*size[0]*size[1];
	      x = Deltay*size[0]+Deltax;
	      y = Deltay*size[0]+Deltax+size[0]*size[1];
	      z = Deltay*size[0]+Deltax+2*size[0]*size[1];
	      
	      /* Erosion */
	      divx = divextended(B[x],SESphCIELab[xSE]);
	      divy = divphi(B[y],SESphCIELab[ySE]);
	      divz = divtheta(B[z],SESphCIELab[zSE]);
	     
	      if ((divx>SphCIELabOUT[xORIG])||((divx==SphCIELabOUT[xORIG])&&(less(divy, SphCIELabOUT[yORIG], 0)))||((divx==SphCIELabOUT[xORIG])&&(SphCIELabOUT[yORIG]==divy)&&(less(divz, SphCIELabOUT[zORIG], 1))))
	      {
		SphCIELabOUT[xORIG] = divx;
		SphCIELabOUT[yORIG] = divy;
		SphCIELabOUT[zORIG] = divz;
	      }
	    }
	  }
	}
      }
    }
    
    /* Convert image in SphCIELab to RGB */
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	
	/* Convert image in SphCIELab to RGB */
	fnlx = SphCIELabOUT[x]*sin(SphCIELabOUT[y])+Ref[0];
	fnly = SphCIELabOUT[x]*cos(SphCIELabOUT[y])*cos(SphCIELabOUT[z])+Ref[1];
	fnlz = SphCIELabOUT[x]*cos(SphCIELabOUT[y])*sin(SphCIELabOUT[z])+Ref[2];
	Deltax = wx*(fnlinv((fnlx+16.00)/116.00+fnly/500.00));
	Deltay = wy*fnlinv((fnlx+16.00)/116.00);
	Deltaz = wz*(fnlinv((fnlx+16.00)/116.00-fnlz/200.00));
	RGBOUT[x] = min(1,max(0,(+3.240479*Deltax-1.537150*Deltay-0.498535*Deltaz)));
	RGBOUT[y] = min(1,max(0,(-0.969256*Deltax+1.875992*Deltay+0.041556*Deltaz)));
	RGBOUT[z] = min(1,max(0,(+0.055648*Deltax-0.204043*Deltay+1.057311*Deltaz)));
      }
    }
  }
  
  /* ################################################################# */
  /* Opening */
  /* ----------------------------------------------------------------- */

  else if(strcmp(input,"open") == 0)
  {
    printf("Spherical CIELab Quantale-Based Opening.\n");
    double *B;
    /* Get the matrix of input Image B in RGB */
    B = mxGetPr(mxCreateNumericArray(size[2], size, mxDOUBLE_CLASS, mxREAL));
    
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	fnlx = fnl((0.412453*A[x]+0.357580*A[y]+0.180423*A[z])/wx);
	fnly = fnl((0.212671*A[x]+0.715160*A[y]+0.072169*A[z])/wy);
	fnlz = fnl((0.019334*A[x]+0.119193*A[y]+0.950227*A[z])/wz);
	Deltax = (116.00*fnly-16.00-Ref[0]);
	Deltay = (500.00*(fnlx-fnly)-Ref[1]);
	Deltaz = (200.00*(fnly-fnlz)-Ref[2]);
	/* SphCIELabOUT will be the least element for Dilation */
	SphCIELabOUT[x] = BOTX;
	SphCIELabOUT[y] = BOTY;
	SphCIELabOUT[z] = BOTZ;
	/* B will be the identity element for Erosion */
	B[x] = EX;
	B[y] = EY;
	B[z] = EZ;
	/* RGBOUT will be the image A in Spherical Coordenates */
	RGBOUT[x] = sqrt(Deltax*Deltax+Deltay*Deltay+Deltaz*Deltaz); /* r */
	RGBOUT[y] = atan2(Deltax,sqrt(Deltay*Deltay+Deltaz*Deltaz)); /* phi */
	RGBOUT[z] = atan2(Deltaz,Deltay); /* theta */
      }
    }
    /* Get Size of Input Matrix */
    sizeSE = mxGetDimensions(prhs[2]);

    /* Get the matrix of input SE */
    SESphCIELab = mxGetPr(prhs[2]);

    int k, l, xSE, ySE, zSE, limX, limY, xORIG, yORIG, zORIG;
    double mult, infy, infz, divx, divy, divz;

    /* Limits of SE */
    limX = floor(sizeSE[0]/2);
    limY = floor(sizeSE[1]/2);

    /* Erosion */
    for (k = 0; k < size[0]; k++)
    {
      for (l = 0; l < size[1]; l++)
      {
	for (i = 0; i < sizeSE[0]; i++)
	{
	  for (j = 0; j < sizeSE[1]; j++)
	  {
	    /* Test if SE fits in image A */
	    if (((k+i-limX)>=0)&&((k+i-limX)<size[0])&&((l+j-limY)>=0)&&((l+j-limY)<size[1]))
	    {
	      xSE = i+j*sizeSE[0];
	      ySE = i+j*sizeSE[0]+sizeSE[0]*sizeSE[1];
	      zSE = i+j*sizeSE[0]+2*sizeSE[0]*sizeSE[1];
	      Deltax = k + (i-limX);
	      Deltay = l + (j-limY);
	      xORIG = l*size[0]+k;
	      yORIG = l*size[0]+k+size[0]*size[1];
	      zORIG = l*size[0]+k+2*size[0]*size[1];
	      x = Deltay*size[0]+Deltax;
	      y = Deltay*size[0]+Deltax+size[0]*size[1];
	      z = Deltay*size[0]+Deltax+2*size[0]*size[1];

	      /* Erosion */
	      divx = divextended(RGBOUT[x],SESphCIELab[xSE]);
	      divy = divphi(RGBOUT[y],SESphCIELab[ySE]);
	      divz = divtheta(RGBOUT[z],SESphCIELab[zSE]);
	     
	      if ((divx>B[xORIG])||((divx==B[xORIG])&&(less(divy, B[yORIG], 0)))||((divx==B[xORIG])&&(B[yORIG]==divy)&&(less(divz, B[zORIG], 1))))
	      {
		B[xORIG] = divx;
		B[yORIG] = divy;
		B[zORIG] = divz;
	      }
	    }
	  }
	}
      }
    }

    /* Dilation */
    for (k = 0; k < size[0]; k++)
    {
      for (l = 0; l < size[1]; l++)
      {
	for (i = 0; i < sizeSE[0]; i++)
	{
	  for (j = 0; j < sizeSE[1]; j++)
	  {
	    /* Test if SE fits in image A */
	    if (((k+i-limX)>=0)&&((k+i-limX)<size[0])&&((l+j-limY)>=0)&&((l+j-limY)<size[1]))
	    {
	      xSE = i+j*sizeSE[0];
	      ySE = i+j*sizeSE[0]+sizeSE[0]*sizeSE[1];
	      zSE = i+j*sizeSE[0]+2*sizeSE[0]*sizeSE[1];
	      Deltax = k + (i-limX);
	      Deltay = l + (j-limY);
	      xORIG = l*size[0]+k;
	      yORIG = l*size[0]+k+size[0]*size[1];
	      zORIG = l*size[0]+k+2*size[0]*size[1];
	      x = Deltay*size[0]+Deltax;
	      y = Deltay*size[0]+Deltax+size[0]*size[1];
	      z = Deltay*size[0]+Deltax+2*size[0]*size[1];
	      
	      /* Dilation */
	      /* Note that if RGBOUT[x] or SESphCIELab[xSE] are
	       * greater or equal to BOTX, then SphCIELabOUT[xORIG]
	       * cannot be greater than it. */
	      if ((B[x]>=BOTX)||(SESphCIELab[xSE]>=BOTX))
		mult = BOTX;
	      else
	      {
		mult = B[x]*SESphCIELab[xSE];
		infy = inf(B[y],SESphCIELab[ySE]);
		infz = inf(B[z],SESphCIELab[zSE]);

		/* Test the greast value between SphCIELabOUT and RGBOUT */
		if ((mult<SphCIELabOUT[xORIG])||((mult==SphCIELabOUT[xORIG])&&(less(SphCIELabOUT[yORIG],infy, 0)))||((mult==SphCIELabOUT[xORIG])&&(SphCIELabOUT[yORIG]==infy)&&(less(SphCIELabOUT[zORIG],infz, 1))))
		{
		    SphCIELabOUT[xORIG] = mult;
		    SphCIELabOUT[yORIG] = infy;
		    SphCIELabOUT[zORIG] = infz;
		}
	      }
	    }
	  }
	}
      }
    }
    
    /* Convert image in SphCIELab to RGB */
    for (i = 0; i < size[0]; i++)
    {
      for (j = 0; j < size[1]; j++)
      {
	x = j*size[0]+i;
	y = j*size[0]+i+size[0]*size[1];
	z = j*size[0]+i+2*size[0]*size[1];
	
	/* Convert image in SphCIELab to RGB */
	fnlx = SphCIELabOUT[x]*sin(SphCIELabOUT[y])+Ref[0];
	fnly = SphCIELabOUT[x]*cos(SphCIELabOUT[y])*cos(SphCIELabOUT[z])+Ref[1];
	fnlz = SphCIELabOUT[x]*cos(SphCIELabOUT[y])*sin(SphCIELabOUT[z])+Ref[2];
	Deltax = wx*(fnlinv((fnlx+16.00)/116.00+fnly/500.00));
	Deltay = wy*fnlinv((fnlx+16.00)/116.00);
	Deltaz = wz*(fnlinv((fnlx+16.00)/116.00-fnlz/200.00));
	RGBOUT[x] = min(1,max(0,(+3.240479*Deltax-1.537150*Deltay-0.498535*Deltaz)));
	RGBOUT[y] = min(1,max(0,(-0.969256*Deltax+1.875992*Deltay+0.041556*Deltaz)));
	RGBOUT[z] = min(1,max(0,(+0.055648*Deltax-0.204043*Deltay+1.057311*Deltaz)));
      }
    }
  }
  /* ################################################################# */
}
/* ################################################################# */
